﻿#synthDrivers/multilang.py
#Copyright (C) 2009-2010 Aleksey Sadovoy <lex@progger.ru>
#This file is covered by the GNU General Public License.

from collections import OrderedDict
from cStringIO import StringIO
import re
import os
from configobj import ConfigObj
from validate import Validator
import config
import globalVars
from synthDriverHandler import SynthDriver,VoiceInfo,_getSynthDriver
from logHandler import log
import queueHandler

val = Validator()
conf=None
multilangConf=None

confspec = ConfigObj(StringIO(
"""# Multilang Synthesizer Configuration File

[multilang]
	#Multilang profile
	[[__many__]]
		defaultLang = string(default='en')
		switchLanguages = boolean(default=true)
		#Language options
		[[[__many__]]]
			synthesizer=string(default="espeak")
			voice=string()
			variant=string()
"""
), list_values=False, encoding="UTF-8")
confspec.newlines = "\r\n"
defaultConfiguration="""# Multilang Synthesizer Configuration File

[multilang]
	[[default]]
		#Производить переключение языков. допустимые значения: true, false 
		switchLanguages = True
		#Язык по-умолчанию. Допустимые значения: en, ru
		defaultLang = ru
		[[[ru]]]
			#используемый синтезатор
			synthesizer = newfon
		[[[en]]]
			synthesizer = espeak
			#Используемый голос. допустимые значения зависят от конкретного синтезатора. см. параметр "voice" в файле конфигурации NVDA конкретного синтезатора
			voice = en-us
"""

def loadConfig():
	"""Loads the configuration from the configFile.
	"""
	global conf,multilangConf
	configFileName=os.path.join(globalVars.appArgs.configPath,"multilang.ini")
	if not os.path.exists(configFileName):
		with open(configFileName,"w") as f:
			f.write(defaultConfiguration)
	conf = ConfigObj(configFileName, configspec = confspec, indent_type = "\t", encoding="UTF-8")
	# Python converts \r\n to \n when reading files in Windows, so ConfigObj can't determine the true line ending.
	conf.newlines = "\r\n"
	conf.validate(val)
	multilangConf=conf["multilang"]

def saveConfig():
	"""Saves the configuration to the config file.
	"""
	global conf
	if not os.path.isdir(globalVars.appArgs.configPath):
		try:
			os.makedirs(globalVars.appArgs.configPath)
		except OSError, e:
			log.warning("Could not create configuration directory")
			log.debugWarning("", exc_info=True)
	try:
		# Copy default settings and formatting.
		conf.validate(val, copy = True)
		conf.write()
		log.info("Multilang configuration saved")
	except Exception, e:
		log.warning("Could not save multilang configuration - probably read only file system")
		log.debugWarning("", exc_info=True)

class Language(object):
	short = ""
	name = ""
	alphabetRegex=None
	defaultSynth="espeak"

	def recognize(self,str):
		if self.alphabetRegex is None: raise NotImplementedError
		return self.alphabetRegex.search(str)!=None

class RussianLanguage(Language):
	short = "ru"
	name = _("russian")
	alphabetRegex = re.compile(U"[а-яё]",re.U|re.I)
	defaultSynth="newfon"

class EnglishLanguage(Language):
	short="en"
	name=_("english")
	alphabetRegex=re.compile(U"[a-z]",re.I|re.U)

#with detection order
languagesList = (EnglishLanguage(),RussianLanguage())
languages = {}
for l in languagesList: languages[l.short] = l

def loadSynthSettings(synth,section):
	validator=Validator()
	realConf=config.conf["speech"][synth.name]
	if synth.isSupported("voice"):
		voice=section["voice"] if "voice" in section and section["voice"] else realConf["voice"]
		try:
			synth.voice = voice
			realConf.configspec=synth.getConfigSpec()
			config.conf.validate(config.val, copy = True,section = realConf)
			#config.conf.validate(config.val, copy = True,section = section)
		except LookupError:
			log.warning("No such voice: %s" % voice)
			# Update the configuration with the correct voice.
			realConf["voice"]=synth.voice
			section["voice"]=synth.voice
	for s in synth.supportedSettings:
		if s.name=="voice": continue
		v=None
		if s.name in section:
			v=validator.check(s.configSpec, section[s.name])
		elif s.name in realConf:
			v=realConf[s.name]
		if v:
			setattr(synth, s.name, v)

def loadSynth(section):
	name = section['synthesizer']
	if name=='auto':
		name='espeak'
	try:
		newSynth=_getSynthDriver(name)()
		updatedConfig=config.updateSynthConfig(newSynth)
		if not updatedConfig:
			loadSynthSettings(newSynth,section)
		else:
			newSynth.saveSettings() #save defaults
		return newSynth
	except:
		#newSynth.terminate()
		raise

class SynthDriver(SynthDriver):
	name = "multilang"
	description = _("Virtual synthesizer which can determine language of text to speak and use appropriate synthesizer")
	supportedSettings=(SynthDriver.VoiceSetting(),)
	_profile = "default"
	talkers=[]
	languages = []
	curLang=0
	profileConf = None
	generatorID=None
	queue=[]
	#-1 - state_terminating
	#0 - state_sleeping
	#1 - state_speechRequest
	#2 - state_speechStarted
	#3 - state_speech
	state=0
	endIndex=0xfffe
	delims=r"((-+(\s|$))|[^\w\-])*"
	splitExpr=re.compile(r"(\w+(\W*\d+\W*)?\-?\w*?[^\S\t]*)*" + delims, re.I|re.U)
	@classmethod
	def check(cls):
		return True

	def __init__(self):
		global multilangConf
		loadConfig()
		self.availableVoices = OrderedDict([(s,VoiceInfo(s,s)) for s in multilangConf.sections])
		self.generatorID=queueHandler.registerGeneratorObject(self.proceedSpeech())

	def terminate(self):
		queueHandler.cancelGeneratorObject(self.generatorID)
		self.terminateSynths()
		#saveConfig()

	def recognizeLanguage(self,string):
		for i in range(len(self.languages)):
			if self.languages[i].recognize(string):
				return i
		return self.profileConf.sections.index(self.profileConf["defaultLang"])

	def terminateSynths(self):
		for synth in self.talkers: synth.terminate()
		self.talkers = []

	def processPortion(self):
		"""Takes string from queue and sends it to the appropriate synthesizer."""
		(string,index,lang) = self.queue.pop(0)
		self.curLang=lang
		self.talkers[self.curLang].speakText(string,index)
		self.lastIndex=index
		self.talkers[self.curLang].speakText(" ",SynthDriver.endIndex)

	def _get_voice(self):
		return self. _profile

	def _set_voice(self,value):
		global multilangConf,languages
		self.terminateSynths()
		self._profile=value
		self.profileConf=multilangConf[value]
		self.languages = [languages[l] for l in self.profileConf.sections]
		self.talkers = [loadSynth(self.profileConf[lang.short]) for lang in self.languages]
		self.curLang =self.profileConf.sections.index(self.profileConf["defaultLang"])

	def proceedSpeech(self):
		while self.state>-1: #not state_terminating
			while self.state == 0: yield #sleeping
			if self.state == 1: #speech request
				if self.queue!=[]: 
					self.processPortion()
					self.state = 2 if self.talkers[self.curLang].name != "newfon" else 3 #speech started
				else:
					self.state = 0 #sleep
			elif self.state == 2: #speech started
				if self.talkers[self.curLang].lastIndex != SynthDriver.endIndex:
					self.state = 3 #speech proceeding
			elif self.state == 3: #speech proceeding
				if self.talkers[self.curLang].lastIndex==SynthDriver.endIndex:
					self.state = 1 #speech request
			yield

	def cancel(self):
		self.talkers[self.curLang].cancel()
		del self.queue[:]
		self.state = 0 #sleep

	def pause(self,switch):
		self.talkers[self.curLang].pause(switch)

	def speakText(self,text,index=None):
		if not self.profileConf["switchLanguages"]:
			self.talkers[self.curLang].speakText(text,index)
			return
		lastLang=None
		for match in SynthDriver.splitExpr.finditer(text):
			if match.group() == "": continue
			lang = self.recognizeLanguage(match.group())
			if lang==lastLang:
				self.queue[-1][0]+=match.group()
			else:
				self.queue.append([match.group(),index,lang])
				lastLang=lang
		if self.state == 0:
			self.state = 1 #speech request